# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""Frame extraction UDF for blob column testing in materialized views."""

from geneva.udfs.video.frame_extractor import ExtractFirstFrame

__all__ = ["ExtractFirstFrame"]
