# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""
Direct-connect integration tests.

Tests Geneva's ability to connect to pre-existing Ray clusters
without managing cluster lifecycle - mimicking customer deployment pattern.
"""
